/* This is a driver for a Logitech Driving Force  USB wheel 
 *
 *    Philip Brown
 *
 *  @(#) dfwheel.c 1.4@(#)
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>

/* We dont actually use the hid routines.. we use the defines */
/* (and that was for wacom. We might not need it for this wheel) */
#include <sys/usb/clients/hid/hid.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/stream.h>
#include <stropts.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


/* This value goes in the USB "policy" info */
#define DFWHEEL_MAX_TRANSFER 10
#define DFWHEEL_BUFFER_SIZE 2048


/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"dfwheel"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	dfwheel_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	dfwheel_probe(dev_info_t *dip);
static	int	dfwheel_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	dfwheel_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	dfwheel_print(dev_t, char*);

static	int	dfwheel_read(dev_t dev, struct uio* ustr, cred_t * credp);

static	int	dfwheel_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	dfwheel_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	dfwheel_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	dfwheel_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	dfwheel_cb_ops = {
	dfwheel_open,		/* open() */
	dfwheel_close,		/* close() */
	nodev,			/* strategy()		*/
	dfwheel_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	dfwheel_read,		/* read() */
	dfwheel_write,		/* write() */
	dfwheel_ioctl,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops dfwheel_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	dfwheel_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	dfwheel_probe,		/* devo_probe */
	dfwheel_attach,		/* devo_attach */
	dfwheel_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&dfwheel_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"dfwheel driver",
	&dfwheel_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   dfwheel_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * dfwheel_glob_soft_handle;

/* This gets filled out on a per-instance basis by dfwheel_attach */
struct dfwheel_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex;
	kcondvar_t read_cv;
	
	usb_config_descr_t usb_config;
	usb_endpoint_descr_t endpoint1,endpoint2;
	int usb_interfacenum;
	usb_interface_descr_t usb_interface_d;
	int openflags; /* tell if we have been open()ed or not */
	usb_pipe_handle_t usb_pipe_handle1,usb_pipe_handle2;

	/* since we are not a streams driver we ahve to make our
	 * own buffer/queue
	 * "end" is where to add new data. "start" is where read()
	 * should take data from
	 */
	unsigned char *readbuffer;
	unsigned char *readbuf_start,*readbuf_end;

};

/* defines used on the "openflags" var in softstate*/
#define DFWHEEL_OPEN_HELD		0x01
#define DFWHEEL_OPEN_SHUTDOWN	0x02 /* flag "shutting down" case */


/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
dfwheel_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}

/* This routine can be called by dfwheel_attach (if errors occur)
 * and dfwheel_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void dfwheel_cleanup(struct dfwheel_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: dfwheel_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if(softstate->readbuffer!=NULL){
		kmem_free(softstate->readbuffer,DFWHEEL_BUFFER_SIZE);
		mutex_destroy(&softstate->instance_mutex);
		cv_destroy(&softstate->read_cv);
	}

	ddi_soft_state_free(dfwheel_glob_soft_handle,instance);
}



/*Like the dfwheel_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See dfwheel_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	dfwheel_debugprint("dfwheel: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&dfwheel_global_mutex, "dfwheel: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct dfwheel_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&dfwheel_glob_soft_handle,
				sizeof(struct dfwheel_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&dfwheel_global_mutex);
		ddi_soft_state_fini(&dfwheel_glob_soft_handle);
		cmn_err(CE_WARN, "dfwheel: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	dfwheel_debugprint("dfwheel: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	dfwheel_debugprint("dfwheel: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	dfwheel_debugprint("dfwheel: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&dfwheel_glob_soft_handle);
		mutex_destroy(&dfwheel_global_mutex);
	}
	return (ret);
}

/************************************************************/


/**********************************************************************/
/*  Our read buffer handling section                                 */
/**********************************************************************/

/* Returns number of bytes in the read buffer.
 * Dont expect a sensible return unless you are holding the
 * instance_mutex.
 */
static int read_buffer_bytes(struct dfwheel_softstate *softstate){
	int datalen;


	if(softstate->readbuf_start > softstate->readbuf_end){
		datalen=(&softstate->readbuffer[DFWHEEL_BUFFER_SIZE]) -
			softstate->readbuf_start;

		datalen+= softstate->readbuf_end - softstate->readbuffer;
	} else {
		datalen=softstate->readbuf_end - softstate->readbuf_start;
	}
	

	return datalen;
}

/*
 * Note that we have a circular buffer, and we do NOT overwrite.
 * After buffer is full, we drop further packets.
 *
 * YOU MUST HAVE instance_mutex BEFORE CALLING!!!
 */
static void
add_to_read_buffer(struct dfwheel_softstate *softstate, char *data, int len){
	unsigned char *bufstart, *bufend;
	unsigned char *bufdatastart,*bufdataend;
	int copylen, bytesfree;

	if(softstate==NULL) {
		dfwheel_debugprint("add_to_read_buffer: softstate==null");
		return;
	}
	if(data==NULL) {
		dfwheel_debugprint("add_to_read_buffer: data==null");
		return;
	}
	bufstart=softstate->readbuffer;
	bufend=&softstate->readbuffer[DFWHEEL_BUFFER_SIZE];
	bufdatastart=softstate->readbuf_start;
	bufdataend=softstate->readbuf_end;


	bytesfree=DFWHEEL_BUFFER_SIZE - read_buffer_bytes(softstate);
	if(bytesfree==0){
		dfwheel_debugprint("add_to_read_buffer: buffer full\n");
		/* silently discard */
		return;
	}
	if(len>bytesfree){
		copylen=bytesfree;
		len=bytesfree;
		dfwheel_debugprint("Note: add_to_read_buffer: partial add\n");
	} else {
		copylen=len;
	}

	if(&bufdataend[copylen]  >= &softstate->readbuffer[DFWHEEL_BUFFER_SIZE]){
		copylen=(&softstate->readbuffer[DFWHEEL_BUFFER_SIZE]) -
			softstate->readbuf_end;
	}
	bcopy(data,bufdataend, copylen);
	bufdataend += copylen;
	if(bufdataend==bufend){
		bufdataend=bufstart;
	}
	len-= copylen;
	if(len>0){
		/* should only be here if we had to wrap around end of buffer*/
		copylen=len;
		bcopy(data,bufdataend, copylen);
		bufdataend += copylen;
		len -= copylen;
	}
	
	softstate->readbuf_end=bufdataend;
	
	if(len!=0){
		cmn_err(CE_WARN, "dfwheel:add_to_read_buffer: internal error\n");
	}

} /* end of add_to_read_buffer() */

/* Make sure you have the instance_mutex before calling this.
 * Returns internal pointers to "clear" the buffer.
 */
static void read_buffer_reset(struct dfwheel_softstate *softstate){
	softstate->readbuf_start=softstate->readbuffer;
	softstate->readbuf_end=softstate->readbuffer;
}

/**********************************************************************/
/* This is probably interrupt context. 
 * mblk_t is a STREAMS style Message Block. Arrg.
 */
static int
dfwheel_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	struct dfwheel_softstate *softstate=(struct dfwheel_softstate*)arg;
	unsigned char*chardata;
	int datalen;

	if(softstate==NULL){
		dfwheel_debugprint("paranoid pays off: pipe_callback NULL arg\n");
		return USB_SUCCESS;
	}
	if(pipe == softstate->usb_pipe_handle1){
		dfwheel_debugprint("pipe1 got callback");
	}
	if(pipe == softstate->usb_pipe_handle2){
		dfwheel_debugprint("pipe2 got callback");
	}
	
	dfwheel_debugprint("dfwheel_pipe_callback called\n");


	if(mblkptr==NULL){
		printf("mblk==NULL\n");
		return (USB_SUCCESS);
	}

	mutex_enter(&softstate->instance_mutex);

	/* If there is data, it is usually in 5-byte blocks */
	datalen=mblkptr->b_wptr - mblkptr->b_rptr;
	
	dfwheel_debugprint(" theoretical len of data is %d\n",datalen);
	

	chardata=mblkptr->b_rptr;

	add_to_read_buffer(softstate,chardata, datalen);

	freemsg(mblkptr);
	cv_signal(&softstate->read_cv);
	mutex_exit(&softstate->instance_mutex);

	return (USB_SUCCESS);
}

/* temp test routine, that is only used for test purposes, essentially.
 */
static void
dfwheel_test_callback(usb_opaque_t arg,int arg1, uint_t arg2)
{
	dfwheel_debugprint("dfwheel_TEST_callback called\n");
	return;
}

/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static int
dfwheel_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{
	/*struct dfwheel_softstate *softstate=(struct dfwheel_softstate*)arg;*/

	dfwheel_debugprint("dfwheel_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	return (USB_SUCCESS);
}

/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS to use.
 */
static int dfwheel_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int dfwheel_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct dfwheel_softstate *softstate;
	int returnval=DDI_FAILURE;
	int instance=getminor((dev_t)arg);
	
	dfwheel_debugprint("dfwheel: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(dfwheel_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static int dfwheel_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	dfwheel_debugprint("dfwheel: probe called\n");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
	dfwheel_debugprint("probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}


#define REQUEST_TYPE USB_DEV_REQ_RECIPIENT_INTERFACE|USB_DEV_REQ_TYPE_CLASS
#define REQUEST_INDEX softstate->usb_interface_d.bInterfaceNumber
#define REQUEST_CMD USB_REQ_SET_CONFIGURATION


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in dfwheel_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  dfwheel_detach()
 */
static int dfwheel_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct dfwheel_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 0);

	dfwheel_debugprint("dfwheel: dfwheel_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "dfwheel: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "dfwheel: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "dfwheel: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	dfwheel_debugprint("dfwheel: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(dfwheel_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "dfwheel: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(dfwheel_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "dfwheel: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(dfwheel_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	


	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);
	cv_init(&softstate->read_cv, NULL, CV_DRIVER,NULL);

	softstate->readbuffer=kmem_zalloc(DFWHEEL_BUFFER_SIZE,KM_SLEEP);
	/* the buffer pointers get cleared in dfwheel_open() */


	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		dfwheel_cleanup(softstate);
		return DDI_FAILURE;
	}
	dfwheel_debugprint("USB config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	dfwheel_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		dfwheel_debugprint("  (resetting interface num to 0)\n");
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		dfwheel_cleanup(softstate);
		return DDI_FAILURE;
	}

	dfwheel_debugprint("number of endpoints = %d\n",
			 softstate->usb_interface_d.bNumEndpoints);

	/* Fill out "endpoint" descriptors */
	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					0,/*endpoint index num*/
					&softstate->endpoint1,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"parse of endpoint1 failed\n");
		dfwheel_cleanup(softstate);
		return DDI_FAILURE;
	}
	printf(" Endpoint1: %x %x %x %x %x %x\n",
	       softstate->endpoint1.bLength,
	       softstate->endpoint1.bDescriptorType,
	       softstate->endpoint1.bEndpointAddress,
	       softstate->endpoint1.bmAttributes,
	       softstate->endpoint1.wMaxPacketSize,
	       softstate->endpoint1.bInterval);

	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					1,/*endpoint index num*/
					&softstate->endpoint2,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"parse of endpoint2 failed\n");
		dfwheel_cleanup(softstate);
		return DDI_FAILURE;
	}
	printf(" Endpoint2: %x %x %x %x %x %x\n",
	       softstate->endpoint2.bLength,
	       softstate->endpoint2.bDescriptorType,
	       softstate->endpoint2.bEndpointAddress,
	       softstate->endpoint2.bmAttributes,
	       softstate->endpoint2.wMaxPacketSize,
	       softstate->endpoint2.bInterval);

	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */

	/* XXX I should really check the dev type, and change the
	 * name based on that stuff
	 */
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* There are no PCI-style interrupts, since we dont get to directly
	 * connect to a device.
	 * but an interrupt handler of sorts gets set up in open()
	 */


	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}


/* used by read() to figure out how many bytes of data
 * the user program wants right now
 */
static int count_uio_size(struct uio* uiostruct){
	int datalen=0;
	int vec_count;
	if(uiostruct==NULL){
		return 0;
	}
	for(vec_count=0;vec_count<uiostruct->uio_iovcnt;vec_count++){
		datalen+=uiostruct->uio_iov[vec_count].iov_len;
	}

	return datalen;
}


/* return 0 on success, or errno */
static int dfwheel_read(dev_t dev, struct uio* ustr, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	struct dfwheel_softstate *softstate;
	int copylen,readbytes;

	if(debug_level>1){
		dfwheel_debugprint("dfwheel: read called\n");
	}

	readbytes=count_uio_size(ustr); /* how many bytes do they WANT?*/
	if(readbytes >=DFWHEEL_BUFFER_SIZE){
		cmn_err(CE_WARN,
			"dfwheel_read: read size must be less than %d\n",
			DFWHEEL_BUFFER_SIZE);
		return EFAULT;/* is this the best errno to return? */
	}

	softstate=ddi_get_soft_state(dfwheel_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"dfwheel_read failed: cannot get softstate?");
		return DDI_FAILURE;
	}
	
	mutex_enter(&softstate->instance_mutex);

	/* Have to return whole read at once. Therefore, have to
	 * wait until read buffer fills up to required amount.
	 */
	while(read_buffer_bytes(softstate) <readbytes){
		if(cv_wait_sig(&softstate->read_cv,&softstate->instance_mutex)<=0){
			mutex_exit(&softstate->instance_mutex);
			return (EIO);
		}
		if(softstate->openflags & DFWHEEL_OPEN_SHUTDOWN){
			dfwheel_debugprint("dfwheel_read: shutdown detected\n");
			mutex_exit(&softstate->instance_mutex);
			return (ENXIO);
		}
	}

	if(debug_level>1){
		dfwheel_debugprint("read(): after cv_wait, bytes in readbuff==%d\n",
			read_buffer_bytes(softstate));
	}

	/*******************************************************************/
	/* This section might be better off as a modular readbuffer routine*/
	/*******************************************************************/

	/* set copylen =  "how many bytes can we do in a single copy?" */
	if(softstate->readbuf_end < softstate->readbuf_start){
		copylen=(&softstate->readbuffer[DFWHEEL_BUFFER_SIZE]) -
			softstate->readbuf_start;
	}else {
		copylen=softstate->readbuf_end - softstate->readbuf_start;
	}
	if(copylen>readbytes){
		copylen=readbytes;
	}

	if(uiomove(softstate->readbuf_start, copylen,UIO_READ,ustr)!=0){
		dfwheel_debugprint("dfwheel_read: uiomove failed?\n");
		mutex_exit(&softstate->instance_mutex);
		return EFAULT;
	}
	softstate->readbuf_start += copylen;
	if(softstate->readbuf_start ==
	   &softstate->readbuffer[DFWHEEL_BUFFER_SIZE]) {
		softstate->readbuf_start=softstate->readbuffer;
	}
	if(softstate->readbuf_start >
	   &softstate->readbuffer[DFWHEEL_BUFFER_SIZE]) {
		cmn_err(CE_WARN,"ERROR: dfwheel_read: internal bufptr error !!");
		softstate->readbuf_start=softstate->readbuffer;
	}

	readbytes -= copylen;

	if(readbytes!=0){
		/* If we are here, it is assumed that we had to wrap around
		 * the end of the buffer, and readbuf_start now points
		 * to the beginning of the whole buffer.
		 */

		copylen=readbytes;

		/* Although undocumented, empirical tests demonstrate that
		 * uiomove recognizes that the read request has been
		 * partially fulfilled, and when called again,
		 * does the second copy AFTER the previously copied data.
		 */
		if(uiomove(softstate->readbuf_start, copylen,UIO_READ,ustr)!=0){
			dfwheel_debugprint("dfwheel_read: uiomove2 failed?\n");
			mutex_exit(&softstate->instance_mutex);
			return EFAULT;
		}
		softstate->readbuf_start+=copylen;
	}
	/**************************************************/
	/*       End of readbuffer section                */
	/**************************************************/
	
	
	
	mutex_exit(&softstate->instance_mutex);

	return 0;
}

static int dfwheel_write(dev_t dev, struct uio *uiop, cred_t *credp){
   	dfwheel_debugprint("dfwheel: write called\n");
   	return (ENXIO);
}


static int dfwheel_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	usb_pipe_policy_t policy,*policyptr;
	struct dfwheel_softstate *softstate;


	softstate=ddi_get_soft_state(dfwheel_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

	policyptr=&policy;
	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	policyptr->pp_callback_arg=(void*)softstate;/* same for ALL callbacks*/

	policyptr->pp_callback=dfwheel_pipe_callback;
	policyptr->pp_exception_callback=dfwheel_exception_callback;

	/* we do not set these, because there are no ISOC endpoints */
	/*
	policyptr->pp_isoc_callback=dfwheel_isoc_callback;
	policyptr->pp_isoc_exception_callback=dfwheel_isoc_exception_callback;
	*/

	policyptr->pp_periodic_max_transfer_size=DFWHEEL_MAX_TRANSFER;
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=0){
		dfwheel_debugprint("dfwheel: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}

	read_buffer_reset(softstate);

	if(usb_pipe_open(softstate->dip,&softstate->endpoint1, policyptr,
			 USB_FLAGS_SLEEP,
			 &softstate->usb_pipe_handle1) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open1 failed");
		mutex_exit(&softstate->instance_mutex);
		return (ENXIO);
	}
	if(usb_pipe_open(softstate->dip,&softstate->endpoint2, policyptr,
			 USB_FLAGS_SLEEP,
			 &softstate->usb_pipe_handle2) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open2 failed");
		mutex_exit(&softstate->instance_mutex);
		return (ENXIO);
	}

	softstate->openflags=1;

	/* Waitaminit.. why do I need to set a "policy", if I can
	 * specify a callback thing here??!! sigh.
	 */
	if(usb_pipe_start_polling(softstate->usb_pipe_handle1,
				  USB_FLAGS_SLEEP,
				  dfwheel_test_callback,NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling of pipe1 FAILED");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}
	if(usb_pipe_start_polling(softstate->usb_pipe_handle2,
				  USB_FLAGS_SLEEP,
				  dfwheel_test_callback,NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling of pipe2 FAILED");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}
	mutex_exit(&softstate->instance_mutex);
	

   	return 0;
}

static int dfwheel_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct dfwheel_softstate *softstate;


	softstate=ddi_get_soft_state(dfwheel_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"dfwheel_close failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	dfwheel_debugprint("dfwheel: close called\n");
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags==0){
		cmn_err(CE_WARN,"ERROR: dfwheel_close: pipe not open");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
		
	}

	usb_pipe_stop_polling(softstate->usb_pipe_handle1,
			      USB_FLAGS_SLEEP,NULL,NULL);
	usb_pipe_stop_polling(softstate->usb_pipe_handle2,
			      USB_FLAGS_SLEEP,NULL,NULL);

	if(usb_pipe_close(&softstate->usb_pipe_handle1,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		dfwheel_debugprint("ERROR: close of pipe1 failed?");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}
	if(usb_pipe_close(&softstate->usb_pipe_handle2,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		dfwheel_debugprint("ERROR: close of pipe2 failed?");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}

	softstate->usb_pipe_handle1=NULL;
	softstate->usb_pipe_handle2=NULL;
	softstate->openflags=0;

	/* Okay, a little redundant since we do this in open(), but
	 * thats okay. It makes sure read_buffer_bytes returns 0.
	 */
	read_buffer_reset(softstate);
	
	mutex_exit(&softstate->instance_mutex);
   	return (0);
}

static int dfwheel_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp){
	int instance=DEVT_TO_INST(dev);
	struct dfwheel_softstate *softstate;

	if(debug_level>1){
		dfwheel_debugprint("dfwheel: ioctl called\n");
	}

	softstate=ddi_get_soft_state(dfwheel_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"dfwheel_ioctl failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	if(cmd==I_NREAD){
		/* Okay we arent a STREAMS driver, but this is
		 * very convenient to support!
		 */
		int numbytes=read_buffer_bytes(softstate);
		if(arg==NULL){
			dfwheel_debugprint("I_NREAD passed NULL??\n");
			return EFAULT;
		}
		if(ddi_copyout(&numbytes,(int*)arg,sizeof(int),flags) != 0){
			dfwheel_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;
	}
	dfwheel_debugprint("do not recognize ioctl %d\n",cmd);
   	return (ENXIO);
}

static int dfwheel_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct dfwheel_softstate *softstate;
   	instance=ddi_get_instance(dip);

	dfwheel_debugprint("dfwheel: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(dfwheel_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"dfwheel_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	dfwheel_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}
